#include <EEPROM.h>
#include "bsec.h"

// Use the config string from generic_33v_3s_4d folder for background mode
// Rename the file as bsec_serialized_configurations_iaq.h and bsec_serialized_configurations_iaq.c
// Also change the variable name of the config string inside these files to bsec_config_iaq.Only then config string for background mode will be loaded correctly
// Use the config string from bme688_demo folder for scanning mode,nothing needs to be modified here
// 2 config files bsec_serialized_configurations_iaq.c/h and bsec_serialized_configurations_selectivity.c/h must be placed in src folder before running this code
#include "bsec_serialized_configurations_selectivity.h"
#include "bsec_serialized_configurations_iaq.h"

#define STATE_SAVE_PERIOD	UINT32_C(360 * 60 * 1000) // 360 minutes - 4 times a day

// Helper functions declarations
void checkIaqSensorStatus(void);
void errLeds(void);
void loadState(void);
void updateState(bool);

// Create an object of the class Bsec
Bsec iaqSensor;
uint8_t bsecState[BSEC_MAX_STATE_BLOB_SIZE] = {0};
uint16_t stateUpdateCounter = 0;

String output;
bool forceState = false;
bool highChannel = false;
bool backgroundAwake;
bool scanningAwake;
bool triggerBackground;
bool triggerScanning;
uint8_t scanningModeFlag;
uint8_t scanCycles = 10;
uint8_t scanCycleCounter = 0;
int64_t totalDuration = 0;
int64_t scanRemTime = 0;
bool startOutputFlag = false;

// Entry point for the example
void setup(void)
{
	EEPROM.begin(BSEC_MAX_STATE_BLOB_SIZE + 1); // 1st address for the length
	Serial.begin(115200);
	for (uint8_t i = 0; i < BSEC_MAX_STATE_BLOB_SIZE + 1; i++)
		EEPROM.write(i, 0);

	/* Setup button interrupt to trigger scanning mode */
	pinMode(2, INPUT_PULLUP);
	attachInterrupt(digitalPinToInterrupt(2), scanning_mode_button_press, FALLING);

	iaqSensor.begin(BME680_I2C_ADDR_LOW, Wire, highChannel);
	output = "\nBSEC library version " + String(iaqSensor.version.major) + "." + String(iaqSensor.version.minor) + "." + String(iaqSensor.version.major_bugfix) + "." + String(iaqSensor.version.minor_bugfix);
	Serial.println(output);
	
	/* Total Profile Duration function call */
	totalDuration = iaqSensor.totalProfDur();
	
	/* Header */
	output = "Timestamp [ms], raw temperature [°C], pressure [hPa], raw relative humidity [%], gas [Ohm], gas measurement index, IAQ, IAQ accuracy, temperature [°C], relative humidity [%], gas estimate, number of outputs, scanning mode flag, remaining time [ms]";
	Serial.println(output);

	checkIaqSensorStatus(); 

	backgroundAwake = true;
	triggerBackground = true;
	scanningAwake = false;  
}

// Function that is looped forever
void loop(void)
{
	if (backgroundAwake) {	
		if (triggerBackground){
			bsec_virtual_sensor_t sensorList[7] = {
			BSEC_OUTPUT_RAW_TEMPERATURE,
			BSEC_OUTPUT_RAW_PRESSURE,
			BSEC_OUTPUT_RAW_HUMIDITY,
			BSEC_OUTPUT_RAW_GAS,
			BSEC_OUTPUT_IAQ,
			BSEC_OUTPUT_SENSOR_HEAT_COMPENSATED_TEMPERATURE,
			BSEC_OUTPUT_SENSOR_HEAT_COMPENSATED_HUMIDITY	
			};

			iaqSensor.setConfig(bsec_config_iaq);
			checkIaqSensorStatus();
			
			loadState(); 
			
			iaqSensor.updateSubscription(sensorList, 7, BSEC_SAMPLE_RATE_LP);
			checkIaqSensorStatus();			

			// Configure the sensor using forced mode
			iaqSensor.setBme680ConfigForced();		  
			
			scanningModeFlag = 0;
			triggerBackground = false;			
		}		
	
	    if (iaqSensor.runForced()) { // If new data is available
			output = String(long(iaqSensor.getTimeMs()));
			output += ", " + String(iaqSensor.rawTemperature);
			output += ", " + String(iaqSensor.pressure);
			output += ", " + String(iaqSensor.rawHumidity);
			output += ", " + String(iaqSensor.gasResistance);
			output += ", " + String(iaqSensor.gasMeasurementIndex);
			output += ", " + String(iaqSensor.iaqEstimate);
			output += ", " + String(iaqSensor.iaqAccuracy);
			output += ", " + String(iaqSensor.temperature);
			output += ", " + String(iaqSensor.humidity);
			output += ", " + String(iaqSensor.gasEstimate);
			output += ", " + String(iaqSensor.nBsecOutputs);
			output += ", " + String(scanningModeFlag);			
			Serial.println(output);
			forceState = false;
			updateState(forceState);
	    } else {
			checkIaqSensorStatus();
	    }
    }
	
	if (scanningAwake) {	
		if (triggerScanning)	{
			bsec_virtual_sensor_t sensorList[5] = {	
			BSEC_OUTPUT_RAW_TEMPERATURE,
			BSEC_OUTPUT_RAW_PRESSURE,
			BSEC_OUTPUT_RAW_HUMIDITY,
			BSEC_OUTPUT_RAW_GAS,			
			BSEC_OUTPUT_GAS_ESTIMATE
			};
			forceState = true;
			updateState(forceState);
			
			iaqSensor.setConfig(bsec_config_selectivity);
			checkIaqSensorStatus();

			iaqSensor.updateSubscription(sensorList, 5, BSEC_SAMPLE_RATE_HIGH_PERFORMANCE);
			checkIaqSensorStatus();			
			
			// Configure the sensor using parallel mode
			iaqSensor.setBme680ConfigParallel();
						    		
			scanningModeFlag = 1;      
			triggerScanning = false;
		}	
		
		if (iaqSensor.runParallel()) { // If new data is available
    		output = String(long(iaqSensor.getTimeMs()));
			output += ", " + String(iaqSensor.rawTemperature);
			output += ", " + String(iaqSensor.pressure);
			output += ", " + String(iaqSensor.rawHumidity);
			output += ", " + String(iaqSensor.gasResistance);
			output += ", " + String(iaqSensor.gasMeasurementIndex);
			output += ", " + String(iaqSensor.iaqEstimate);
			output += ", " + String(iaqSensor.iaqAccuracy);
			output += ", " + String(iaqSensor.temperature);
			output += ", " + String(iaqSensor.humidity);
			output += ", " + String(iaqSensor.gasEstimate);
			output += ", " + String(iaqSensor.nBsecOutputs);
			output += ", " + String(scanningModeFlag);
			
			if(iaqSensor.gasMeasurementIndex == 0)  {
				scanRemTime = totalDuration - iaqSensor.heatrstepDur[iaqSensor.gasMeasurementIndex];
				if(!startOutputFlag)	{
					startOutputFlag = true;
				}
			}
			else	{
				scanRemTime -= iaqSensor.heatrstepDur[iaqSensor.gasMeasurementIndex];      
			} 
			if(startOutputFlag)  {
				output += ", " + String(long(scanRemTime));
			}
			Serial.println(output); 
			
			if (iaqSensor.gasMeasurementIndex == 9) {
				scanCycleCounter = scanCycleCounter + 1;			    
				if (scanCycleCounter == scanCycles) {					
					backgroundAwake = true;
					triggerBackground = true;
					scanningAwake = false; 
					scanCycleCounter = 0;
				}
			}
		} 
		else {
			checkIaqSensorStatus();
		}
	}	
}

// Helper function definitions
void checkIaqSensorStatus(void)
{
    if (iaqSensor.status != BSEC_OK) {
		if (iaqSensor.status < BSEC_OK) {
		    output = "BSEC error code : " + String(iaqSensor.status);
		    Serial.println(output);
		    for (;;)
			    errLeds(); /* Halt in case of failure */
		} else {
		    output = "BSEC warning code : " + String(iaqSensor.status);
		    Serial.println(output);
		}
    }

    if (iaqSensor.bme680Status != BME680_OK) {
		if (iaqSensor.bme680Status < BME680_OK) {
		    output = "BME680 error code : " + String(iaqSensor.bme680Status);
		    Serial.println(output);
		    for (;;)
				errLeds(); /* Halt in case of failure */
		} else {
		    output = "BME680 warning code : " + String(iaqSensor.bme680Status);
		    Serial.println(output);
		}
    }
}

void errLeds(void)
{
    pinMode(LED_BUILTIN, OUTPUT);
    digitalWrite(LED_BUILTIN, HIGH);
    delay(100);
    digitalWrite(LED_BUILTIN, LOW);
    delay(100);
}

void loadState(void)
{
    if (EEPROM.read(0) == BSEC_MAX_STATE_BLOB_SIZE) {
		// Existing state in EEPROM
		Serial.println("Reading state from EEPROM");

		for (uint8_t i = 0; i < BSEC_MAX_STATE_BLOB_SIZE; i++) {
			bsecState[i] = EEPROM.read(i + 1);
			Serial.println(bsecState[i], HEX);
		}

		iaqSensor.setState(bsecState);
		checkIaqSensorStatus();
    } else {
		// Erase the EEPROM with zeroes
		Serial.println("Erasing EEPROM");

		for (uint8_t i = 0; i < BSEC_MAX_STATE_BLOB_SIZE + 1; i++)
		    EEPROM.write(i, 0);

		EEPROM.commit();
    }
}

void updateState(bool forceState)
{
    bool update = false;
    if (forceState) {
		update = true;    
    }
    else if (stateUpdateCounter == 0) {
		/* First state update when IAQ accuracy is >= 1 */
		if (iaqSensor.iaqAccuracy >= 3) {
			update = true;
			stateUpdateCounter++;
		}
    } else {
		/* Update every STATE_SAVE_PERIOD minutes */
		if ((stateUpdateCounter * STATE_SAVE_PERIOD) < millis()) {
			update = true;
			stateUpdateCounter++;
        }
    }

    if (update) {
		iaqSensor.getState(bsecState);
		checkIaqSensorStatus();

		Serial.println("Writing state to EEPROM");

		for (uint8_t i = 0; i < BSEC_MAX_STATE_BLOB_SIZE ; i++) {
			EEPROM.write(i + 1, bsecState[i]);
			Serial.println(bsecState[i], HEX);
		}

		EEPROM.write(0, BSEC_MAX_STATE_BLOB_SIZE);
		EEPROM.commit();		
    }
}

/*!
   @brief       Interrupt handler for press of a ULP plus button

   @return      none
*/
void scanning_mode_button_press()
{
	/* We trigger scanning mode with the button press from next iteration*/     
    backgroundAwake = false;
	scanningAwake = true; 
    triggerScanning	= true;
}
